#! /usr/bin/env python

import warnings

from pyslet.html401 import *    # noqa

warnings.warn(
    "pyslet.html40_19991224 is deprecated, use pyslet.html401 instead",
    DeprecationWarning, stacklevel=3)
