#! /usr/bin/env python

import warnings
from pyslet.xml.namespace import *      # noqa

warnings.warn(
    "pyslet.xmlnames20091208 is deprecated, use pyslet.xml.namespace instead",
    DeprecationWarning, stacklevel=3)
